"""Logic for compiling requirements"""
from __future__ import print_function

import logging

import pkg_resources

import qer.dists
import qer.metadata
import qer.repos.pypi
import qer.utils
from qer.utils import parse_requirement, merge_requirements
import qer.repos.repository

from qer.dists import DistributionCollection
from qer.repos.repository import NoCandidateException


ROOT_REQ = 'root__a.out'
CONSTRAINTS_REQ = 'constraints__a.out'

BLACKLIST = [
    'setuptools'
]

MAX_DOWNGRADE = 3


def compile_roots(node, source, repo, dists, depth=1, verbose=False):
    """

    Args:
        node (qer.dists.DependencyNode):
        source (DependencyNode):
        repo (Repository):
        dists (DistributionCollection):
        depth:
        verbose:

    Returns:

    """
    logger = logging.getLogger('qer.compile')

    if verbose:
        print(' ' * depth + node.key, end='')

    if node.metadata is not None:
        if verbose:
            print(' ... REUSE')
        logger.info('Reusing dist %s %s', node.metadata.name, node.metadata.version)

        if node.metadata.meta:
            for req in list(node.dependencies):
                compile_roots(req, node, repo, dists, depth=depth + 1, verbose=verbose)
    else:
        spec_req = node.build_constraints()
        first_failure = None
        for attempt in range(MAX_DOWNGRADE):
            dist, cached = repo.get_candidate(spec_req)
            source_repo = repo.source_of(spec_req)

            if verbose:
                if cached:
                    print(' ... CACHED[{}] ({})'.format(dist, source_repo))
                else:
                    print(' ... DOWNLOAD ({})'.format(source_repo))

            nodes_to_recurse = set()
            metadata = None

            try:
                if isinstance(dist, qer.metadata.DistInfo):
                    metadata = dist
                else:
                    metadata = qer.metadata.extract_metadata(dist, origin=source_repo)
                nodes_to_recurse = dists.add_dist(metadata, source, source.dependencies[node])
                if nodes_to_recurse:
                    for recurse_node in nodes_to_recurse:
                        for req in list(recurse_node.dependencies):
                            compile_roots(req, recurse_node, repo, dists, depth=depth + 1, verbose=verbose)
                first_failure = None
                break
            except qer.metadata.MetadataError as meta_error:
                ex = meta_error
                spec_req = merge_requirements(spec_req,
                                              parse_requirement('{}!={}'.format(meta_error.name, meta_error.version)))
            except NoCandidateException as no_candidate_ex:
                ex = no_candidate_ex
                spec_req = merge_requirements(spec_req,
                                              parse_requirement('{}!={}'.format(metadata.name, metadata.version)))

            if first_failure is None:
                first_failure = ex

            for node in nodes_to_recurse:
                dists.remove_dists(node, remove_upstream=False)

        if first_failure is not None:
            raise first_failure


def _generate_constraints(dists):
    for dist in dists:
        if dist.metadata.name in BLACKLIST:
            continue
        if dist.metadata.name.startswith(ROOT_REQ):
            continue

        req = dists.build_constraints(dist.metadata.name)
        if req.specifier:
            yield req


def _build_root_metadata(roots, name):
    return qer.dists.DistInfo(name, '0', roots, meta=True)


def perform_compile(input_reqs, repo, constraint_reqs=None):
    """
    Perform a compilation using the given inputs and constraints
    Args:
        input_reqs (list[pkg_resources.Requirement] or
                    dict[str, list[pkg_resources.Requirement]]):
            List of mapping of input requirements. If provided a mapping,
            requirements will be kept separate during compilation for better
            insight into the resolved requirements
        constraint_reqs (list[pkg_resources.Requirement] or None): Constraints to use
            when compiling
    Returns:
        tuple[DistributionCollection, DistributionCollection, dict]
    """
    results = qer.dists.DistributionCollection()

    if constraint_reqs:
        constraint_node = results.add_dist(qer.dists.RequirementsFile(CONSTRAINTS_REQ, constraint_reqs), None, None)

    nodes = set()
    if isinstance(input_reqs, dict):
        fake_reqs = []
        for idx, req_source in enumerate(input_reqs):
            roots = input_reqs[req_source]
            name = '{}{}'.format(ROOT_REQ, idx)
            fake_reqs.append(pkg_resources.Requirement(name))
            nodes |= results.add_dist(qer.dists.RequirementsFile(req_source, roots), None, None)
    else:
        nodes |= results.add_dist(qer.dists.RequirementsFile(ROOT_REQ, input_reqs), None, None)

    try:
        for node in nodes:
            compile_roots(node, None, repo, dists=results)
    except NoCandidateException as ex:
        ex.results = results
        raise

    if constraint_reqs:
        results.remove_dists(list(constraint_node)[0])

    for node in results:
        if node.metadata is None:
            raise NoCandidateException(node.build_constraints(), results=results)
    return results



import itertools
import os
from collections import defaultdict
import logging

try:
    from functools32 import lru_cache
except ImportError:
    from functools import lru_cache

import pkg_resources


def _req_iter_from_file(reqfile_name):
    with open(reqfile_name, 'r') as reqfile:
        for req_line in reqfile:
            req_line = req_line.strip()
            if not req_line:
                continue
            if req_line.startswith('-r'):
                for req in _req_iter_from_file(os.path.join(
                        os.path.dirname(reqfile_name),
                        req_line.split(' ')[1].strip())):
                    yield req
            elif req_line.startswith('--index-url') or req_line.startswith('--extra-index-url'):
                pass
            elif req_line.startswith('#'):
                pass
            else:
                try:
                    yield parse_requirement(req_line)
                except ValueError:
                    logging.getLogger('req.utils').exception('Failed to parse %s', req_line)
                    raise


def reqs_from_files(requirements_files):
    """
    Args:
        requirements_files (list[str]): Requirements files
    """
    raw_reqs = iter([])
    for reqfile_name in requirements_files:
        raw_reqs = itertools.chain(raw_reqs, _req_iter_from_file(reqfile_name))

    reqs = defaultdict(lambda: None)
    for req in raw_reqs:
        reqs[req.name] = merge_requirements(reqs[req.name], req)

    return list(reqs.values())


# @lru_cache(maxsize=None)
def parse_requirement(req_text):
    if req_text[0] == '#':
        return None
    return pkg_resources.Requirement.parse(req_text)


@lru_cache(maxsize=None)
def parse_version(version):
    """
    Args:
        version (str): Version to parse
    """
    return pkg_resources.parse_version(version)


def parse_requirements(reqs):
    for req in reqs:
        result = parse_requirement(req)
        if result is not None:
            yield result


def merge_extras(extras1, extras2):
    if extras1 is None:
        return extras2
    if extras2 is None:
        return extras1
    return tuple(sorted(list(set(extras1) | set(extras2))))


@lru_cache(maxsize=None)
def merge_requirements(req1, req2):
    if req1 is not None and req2 is None:
        return req1
    if req2 is not None and req1 is None:
        return req2

    assert normalize_project_name(req1.name) == normalize_project_name(req2.name)
    all_specs = set(req1.specs or []) | set(req2.specs or [])

    # Handle markers
    if req1.marker and req2.marker:
        if str(req1.marker) != str(req2.marker):
            if str(req1.marker) in str(req2.marker):
                new_marker = ';' + str(req1.marker)
            elif str(req2.marker) in str(req1.marker):
                new_marker = ';' + str(req2.marker)
            else:
                new_marker = ''
        else:
            new_marker = ';' + str(req1.marker)
    else:
        new_marker = ''

    extras = merge_extras(req1.extras, req2.extras)
    extras_str = ''
    if extras:
        extras_str = '[' + ','.join(extras) + ']'
    req_str = normalize_project_name(req1.name) + extras_str + ','.join(''.join(parts) for parts in all_specs) + new_marker
    return parse_requirement(req_str)


name_cache = {}


def normalize_project_name(project_name):
    if project_name in name_cache:
        return name_cache[project_name]
    else:
        value = project_name.lower().replace('-', '_') #.replace('.', '_')
        name_cache[project_name] = value
        return value


def filter_req(req, extra):
    if extra and not req.marker:
        return False
    keep_req = True
    if req.marker:
        keep_req = req.marker.evaluate({'extra': extra})
    return keep_req


def is_pinned_requirement(req):
    """
    Returns whether an InstallRequirement is a "pinned" requirement.

    An InstallRequirement is considered pinned if:
    - Is not editable
    - It has exactly one specifier
    - That specifier is "=="
    - The version does not contain a wildcard

    Examples:
        django==1.8   # pinned
        django>1.8    # NOT pinned
        django~=1.8   # NOT pinned
        django==1.*   # NOT pinned
    """

    return any((spec.operator == '==' or spec.operator == '===') and not spec.version.endswith('.*')
               for spec in req.specifier)




import argparse
import logging
import os
from collections import defaultdict

from hashlib import sha256

import six

from qer.repos.source import SourceRepository
from qer.utils import merge_requirements
import qer.utils


def run_hash(reqs):
    """
    Args:
        reqs (list[pkg_resources.Requirement]): Requirements files
    """
    reqs = defaultdict(lambda: None)
    for req in reqs:
        reqs[req.name] = merge_requirements(reqs[req.name], req)

    hasher = sha256()
    for req in reqs.values():
        hasher.update(str(req).encode('utf-8'))
    return hasher.hexdigest()


def hash_main():
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('requirement_files', nargs='+',
                        help='Input requirements files')
    parser.add_argument('-e', '--extra', nargs='+', default=[],
                        help='Extras to include for every discovered distribution')

    args = parser.parse_args()

    extras = tuple(args.extra)

    if len(args.requirement_files) == 1 and os.path.isdir(args.requirement_files[0]):
        reqs = []
        source_repo = SourceRepository(args.requirement_files[0])
        for candidates in six.itervalues(source_repo.distributions):
            for candidate in candidates:
                reqs.extend(candidate.preparsed.requires(extras=extras))
    else:
        reqs = qer.utils.reqs_from_files(args.requirement_files)

    print(run_hash(reqs))


if __name__ == '__main__':
    hash_main()



from __future__ import print_function

import logging
import sys
from collections import defaultdict

import six
from six.moves import map as imap

import qer.dists
import qer.utils

from qer import utils

from .repository import Repository, Candidate, DistributionType, RequiresPython


class SolutionRepository(Repository):
    def __init__(self, filename, allow_prerelease=None):
        super(SolutionRepository, self).__init__(allow_prerelease=allow_prerelease)
        self.filename = filename
        self.solution = load_from_file(self.filename)
        self._logger = logging.getLogger('qer.repository.solution')

    def __repr__(self):
        return '--solution {}'.format(self.filename)

    @property
    def logger(self):
        return self._logger

    def get_candidates(self, req):
        try:
            node = self.solution[req.name]
            candidate = Candidate(
                node.key,
                node.metadata,
                node.metadata.version,
                RequiresPython(None),
                'any',
                None,
                DistributionType.SOURCE)
            candidate.preparsed = node.metadata
            return [candidate]
        except KeyError:
            return []

    def resolve_candidate(self, candidate):
        return candidate.filename, True

    def close(self):
        pass


def load_from_file(filename):
    result = qer.dists.DistributionCollection()

    with open(filename) as reqfile:
        for line in reqfile.readlines():
            req_part, _, source_part = line.partition('#')
            req = qer.utils.parse_requirement(req_part)
            source_part = source_part.strip()

            sources = source_part.split(', ')

            pkg_names = imap(lambda x: x.split(' ')[0], sources)
            constraints = imap(lambda x: x.split(' ')[1].replace('(', '').replace(')', '') if '(' in x else None, sources)

            version = qer.utils.parse_version(list(req.specifier)[0].version)
            metadata = qer.dists.DistInfo(req.name, version, [])
            result.add_dist(metadata, None, req)

            for name, constraints in zip(pkg_names, constraints):
                if name and not (name.endswith('.txt') or name.endswith('.out')):
                    constraint_req = qer.utils.parse_requirement(name)
                    result.add_dist(constraint_req.name, None, constraint_req)
                    reverse_dep = result[name]
                else:
                    reverse_dep = None
                result.add_dist(metadata.name,
                                reverse_dep,
                                qer.utils.parse_requirement('{}{}{}'.format(metadata.name,
                                                                            ('[' + ','.join(req.extras) + ']') if req.extras else '',
                                                                            constraints if constraints else '')))

    nodes_to_remove = []
    for node in result:
        if node.metadata is not None:
            try:
                requirements = [value for dep_node, value in six.iteritems(node.dependencies)
                                if dep_node.metadata.name != node.metadata.name]
                if node.extra:
                    requirements = [qer.utils.parse_requirement('{} ; extra=="{}"'.format(req, node.extra))
                                    for req in requirements]
                node.metadata.reqs.extend(requirements)
            except Exception:
                print('Error while processing requirement {}'.format(node), file=sys.stderr)
                raise
        else:
            nodes_to_remove.append(node)

    # for root_req in req_mapping:
    #     for req in req_mapping[root_req]:
    #         req_constraints = req_mapping[root_req][req][0] or ''
    #         if '[' in root_req:
    #             req_name, extras = root_req.split('[')
    #             extras = extras.replace(']', '')
    #             extras = extras.split(',')
    #             for extra in extras:
    #                 result.nodes[qer.utils.normalize_project_name(req_name)].metadata.reqs.append(
    #                     qer.utils.parse_requirement(req + req_constraints + ' ; extra=="{}"'.format(extra)))
    #         else:
    #             result.nodes[qer.utils.normalize_project_name(root_req)].metadata.reqs.append(qer.utils.parse_requirement(req + req_constraints))
    for node in nodes_to_remove:
        try:
            del result.nodes[node.key]
        except KeyError:
            pass
    return result



"""Repository to handle pulling packages from online package indexes"""
import logging
import os
from hashlib import sha256

import requests
from six.moves import html_parser
from six.moves import urllib

from .repository import Repository, process_distribution

try:
    from functools32 import lru_cache
except ImportError:
    from functools import lru_cache


logger = logging.getLogger('qer.pypi')


class LinksHTMLParser(html_parser.HTMLParser):
    def __init__(self, url):
        html_parser.HTMLParser.__init__(self)
        self.url = url
        self.dists = []
        self.active_link = None
        self.active_requires_python = None

    def handle_starttag(self, tag, attrs):
        self.active_link = None
        if tag == 'a':
            self.active_requires_python = None
            for attr in attrs:
                if attr[0] == 'href':
                    self.active_link = self.url, attr[1]
                elif attr[0] == 'metadata-requires-python' or attr[0] == 'data-requires-python':
                    self.active_requires_python = attr[1] or None

    def handle_data(self, data):
        if self.active_link is None:
            return
        candidate = process_distribution(self.active_link, data, self.active_requires_python)
        if candidate is not None:
            self.dists.append(candidate)

    def error(self, message):
        raise RuntimeError(message)


@lru_cache(maxsize=None)
def _scan_page_links(index_url, project_name, session):
    """

    Args:
        index_url:
        project_name:
        session (requests.Session): Session

    Returns:
        (list[Candidate])
    """
    url = '{index_url}/{project_name}'.format(index_url=index_url, project_name=project_name)
    logging.getLogger('qer.net.pypi').info('Fetching versions for %s', project_name)
    if session is None:
        session = requests
    response = session.get(url + '/')

    parser = LinksHTMLParser(response.url)
    parser.feed(response.content.decode('utf-8'))

    return parser.dists


def _do_download(logger, filename, link, session, wheeldir):
    url, link = link
    split_link = link.split('#sha256=')
    sha = split_link[1]

    output_file = os.path.join(wheeldir, filename)

    if os.path.exists(output_file):
        hasher = sha256()
        with open(output_file, 'rb') as handle:
            while True:
                block = handle.read(4096)
                if not block:
                    break
                hasher.update(block)
        if hasher.hexdigest() == sha:
            logger.info('Reusing %s', output_file)
            return output_file, True

    full_link = urllib.parse.urljoin(url, link)
    logger.info('Downloading %s -> %s', full_link, output_file)
    if session is None:
        session = requests
    response = session.get(full_link, stream=True)

    with open(output_file, 'wb') as handle:
        for block in response.iter_content(4 * 1024):
            handle.write(block)
    return output_file, False


class PyPIRepository(Repository):
    def __init__(self, index_url, wheeldir, allow_prerelease=None):
        super(PyPIRepository, self).__init__(allow_prerelease)

        self.index_url = index_url
        self.wheeldir = wheeldir
        self.allow_prerelease = allow_prerelease

        self._logger = logging.getLogger('qer.pypi')
        self.session = requests.Session()

    def __repr__(self):
        return '--index-url {}'.format(self.index_url)

    @property
    def logger(self):
        return self._logger

    def get_candidates(self, req):
        return _scan_page_links(self.index_url, req.name, self.session)

    def resolve_candidate(self, candidate):
        return _do_download(self.logger, candidate.filename, candidate.link, self.session, self.wheeldir)

    def close(self):
        self.session.close()



from qer.repos.repository import Repository, BaseRepository, NoCandidateException


class MultiRepository(BaseRepository):
    def __init__(self, *repositories):
        super(MultiRepository, self).__init__()
        self.repositories = list(repositories)  # type: list[Repository]
        self.source = {}

    def get_candidate(self, req):
        last_ex = None
        for repo in self.repositories:
            try:
                candidates = repo.get_candidates(req)
                result = repo._do_get_candidate(req, candidates)
                self.source[req.name] = repo
                return result
            except NoCandidateException as ex:
                last_ex = ex
        raise last_ex

    def get_candidates(self, req):
        candidates = []
        for repo in self.repositories:
            try:
                repo_candidates = repo.get_candidates(req)
                candidates.extend(repo_candidates)
            except NoCandidateException as ex:
                pass
        return candidates

    def source_of(self, req):
        return self.source[req.name]



from __future__ import print_function
import argparse
import shutil
import tempfile

import pkg_resources

from qer.cmdline import add_repo_args, build_repo


def candidates_main():
    parser = argparse.ArgumentParser()
    parser.add_argument('project_name', help='Print candidates found for the project')
    add_repo_args(parser)

    args = parser.parse_args()

    wheeldir = tempfile.mkdtemp()
    repo = build_repo(args.source, None, args.find_links, args.index_url,
                      args.no_index, wheeldir)

    try:
        candidates = repo.get_candidates(pkg_resources.Requirement.parse(args.project_name))
        for candidate in candidates:
            print(candidate)
    finally:
        shutil.rmtree(wheeldir)


if __name__ == '__main__':
    candidates_main()



from __future__ import print_function

import abc
import struct
import enum
import platform
import re
import sys

import pkg_resources
import six

import qer.metadata
import qer.utils

INTERPRETER_TAGS = {
    'CPython': 'cp',
    'IronPython': 'ip',
    'PyPy': 'pp',
    'Jython': 'jy',
}


def _get_platform_tag():
    is_32 = struct.calcsize("P") == 4
    if sys.platform == 'win32':
        if is_32:
            tag = 'win32'
        else:
            tag = 'win_amd64'
    elif sys.platform.startswith('linux'):
        if is_32:
            tag = 'manylinux1_' + platform.machine()
        else:
            tag = 'manylinux1_x86_64'
    else:
        raise ValueError('Unsupported platform: {}'.format(sys.platform))
    return tag


INTERPRETER_TAG = INTERPRETER_TAGS.get(platform.python_implementation(), 'cp')
PY_VERSION_NUM = str(sys.version_info.major) + str(sys.version_info.minor)

PLATFORM_TAG = _get_platform_tag()
EXTENSIONS = ('.whl', '.tar.gz', '.tgz', '.zip')


class DistributionType(enum.Enum):
    SOURCE = 2
    WHEEL = 1
    SDIST = 0


class RequiresPython(object):
    OPS = {
        '<': lambda x, y: x < y,
        '>': lambda x, y: x > y,
        '==': lambda x, y: x == y,
        '!=': lambda x, y: x != y,
        '>=': lambda x, y: x >= y,
        '<=': lambda x, y: x <= y
    }
    SYS_PY_VERSION = pkg_resources.parse_version(sys.version.split(' ')[0])
    SYS_PY_MAJOR = pkg_resources.parse_version('{}'.format(sys.version_info.major))
    SYS_PY_MAJOR_MINOR = pkg_resources.parse_version('{}.{}'.format(sys.version_info.major,
                                                                    sys.version_info.minor))
    WHEEL_VERSION_TAGS = ('py2' if six.PY2 else 'py3',
                          INTERPRETER_TAG + PY_VERSION_NUM,
                          'py' + PY_VERSION_NUM)

    def __init__(self, py_version):
        self.py_version = py_version

    def check_compatibility(self):
        if self.py_version is None or self.py_version == ():
            return True
        if isinstance(self.py_version, tuple):
            return any(version in RequiresPython.WHEEL_VERSION_TAGS
                       for version in self.py_version)

        return all(self._check_py_constraint(part) for part in self.py_version.split(','))

    @classmethod
    def _check_py_constraint(cls, version_constraint):
        ref_version = cls.SYS_PY_VERSION
        version_part = re.split('[!=<>~]', version_constraint)[-1].strip()
        operator = version_constraint.replace(version_part, '').strip()
        if version_part.endswith('.*'):
            version_part = version_part.replace('.*', '')
            dotted_parts = len(version_part.split('.'))
            if dotted_parts == 2:
                ref_version = cls.SYS_PY_MAJOR_MINOR
            if dotted_parts == 1:
                ref_version = cls.SYS_PY_MAJOR
        version = pkg_resources.parse_version(version_part)
        return cls.OPS[operator](ref_version, version)

    @property
    def tag_score(self):
        result = 100
        version_val = None
        if not isinstance(self.py_version, tuple):
            version_val = self.py_version
        elif len(self.py_version) == 1:
            version_val = self.py_version[0]

        if version_val is not None:
            for tag_type in tuple(INTERPRETER_TAGS.values()) + ('py',):
                version_val = version_val.replace(tag_type, '')
            try:
                result += int(version_val)
            except ValueError:
                pass

        return result

    def __eq__(self, other):
        return self.py_version == other.py_version

    def __str__(self):
        if self.py_version is None or self.py_version == ():
            return 'any'

        if isinstance(self.py_version, tuple):
            return ','.join(sorted(self.py_version))
        return ''


class Candidate(object):
    def __init__(self, name, filename, version, py_version, platform, link,
                 candidate_type=DistributionType.SDIST):
        """

        Args:
            name:
            filename:
            version:
            py_version (RequiresPython): Python version
            platform:
            link:
            type:
        """
        self.name = name
        self.filename = filename
        self.version = version
        self.py_version = py_version   # type: RequiresPython
        self.platform = platform
        self.link = link
        self.type = candidate_type

        # Sort based on tags to make sure the most specific distributions
        # are matched first
        # self.sortkey = (candidate_type.value, version, self.tag_score)
        # self.sortkey = (version, self.tag_score)
        self.sortkey = (version, candidate_type.value, self.tag_score)

        self.preparsed = None

    def _calculate_tag_score(self):
        tag_score = self.py_version.tag_score
        if platform != 'any':
            tag_score += 1000

        return tag_score

    @property
    def tag_score(self):
        result = self.py_version.tag_score
        if platform != 'any':
            result += 1000

        return result

    def __eq__(self, other):
        return (self.name == other.name and
                self.filename == other.filename and
                self.version == other.version and
                self.py_version == other.py_version and
                self.platform == other.platform and
                self.link == other.link and
                self.type == other.type)

    def __repr__(self):
        return 'Candidate(name={}, filename={}, version={}, py_version={}, platform={}, link={})'.format(
            self.name, self.filename, self.version, self.py_version, self.platform, self.link
        )

    def __str__(self):
        py_version_str = str(self.py_version) + '-'
        return '{} {}-{}-{}{}'.format(
            self.type.name,
            self.name, self.version, py_version_str, self.platform)


class NoCandidateException(Exception):
    def __init__(self, req, results=None):
        super(NoCandidateException, self).__init__()
        self.req = req
        self.results = results
        self.check_level = 0

    def __str__(self):
        return 'NoCandidateException - no candidate for "{}" satisfies {}'.format(
            self.req.name,
            self.req.specifier
        )


def process_distribution(source, filename, py_version=None):
    candidate = None
    if '.whl' in filename:
        candidate = _wheel_candidate(source, filename)
    elif '.tar.gz' in filename or '.tgz' in filename or '.zip' in filename:
        candidate = _tar_gz_candidate(source, filename, py_version=py_version)
    return candidate


def _wheel_candidate(source, filename):
    data_parts = filename.split('-')
    name = data_parts[0]
    version = pkg_resources.parse_version(data_parts[1])
    plat = data_parts[4].split('.')[0]
    return Candidate(name,
                     filename,
                     version,
                     RequiresPython(tuple(data_parts[2].split('.'))),
                     plat,
                     source,
                     candidate_type=DistributionType.WHEEL)


def _tar_gz_candidate(source, filename, py_version=None):
    name, version = qer.metadata.parse_source_filename(filename)
    return Candidate(name, filename, version, RequiresPython(py_version), 'any',
                     source, candidate_type=DistributionType.SDIST)


def _check_platform_compatibility(py_platform):
    return py_platform == 'any' or py_platform.lower() == PLATFORM_TAG


class BaseRepository(six.with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    def get_candidate(self, req):
        raise NotImplementedError()

    def source_of(self, req):
        return self

    def force_extras(self):
        return ()


class CantUseReason(enum.Enum):
    U_CAN_USE = 0
    WRONG_PYTHON_VERSION = 2
    WRONG_PLATFORM = 3
    IS_PRERELEASE = 4
    VERSION_NO_SATISFY = 5


class Repository(six.with_metaclass(abc.ABCMeta, BaseRepository)):
    def __init__(self, allow_prerelease=None):
        if allow_prerelease is None:
            allow_prerelease = False

        self.allow_prerelease = allow_prerelease

    @abc.abstractmethod
    def get_candidates(self, req):
        """
        Fetch all available candidates for a project_name
        Args:
            req (Requirement): Requirement to get candidates for

        Returns:
            (list[Candidate]) List of candidates
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def resolve_candidate(self, candidate):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def logger(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def close(self):
        raise NotImplementedError()

    def get_candidate(self, req):
        candidates = self.get_candidates(req)
        return self._do_get_candidate(req, candidates)

    def _do_get_candidate(self, req, candidates):
        self.logger.info('Getting candidate for %s', req)
        candidates = self._sort_candidates(candidates)
        has_equality = qer.utils.is_pinned_requirement(req)

        check_level = 1
        for candidate in candidates:
            check_level += 1
            if not candidate.py_version.check_compatibility():
                continue

            check_level += 1
            if not _check_platform_compatibility(candidate.platform):
                continue

            check_level += 1
            if not has_equality and not self.allow_prerelease and candidate.version.is_prerelease:
                continue

            check_level += 1
            if not req.specifier.contains(candidate.version, prereleases=has_equality):
                continue

            check_level += 1
            if candidate.type == DistributionType.SDIST:
                self.logger.warning('Considering source distribution for %s', candidate.name)
            return self.resolve_candidate(candidate)

        ex = NoCandidateException(req)
        ex.check_level = check_level
        raise ex

    def why_cant_I_use(self, req, candidate):
        try:
            self._do_get_candidate(req, (candidate,))
            raise ValueError('This requirement can be used')
        except NoCandidateException as ex:
            return CantUseReason(ex.check_level)

    def _sort_candidates(self, candidates):
        """

        Args:
            candidates:

        Returns:
            (list[Candidate])
        """
        return sorted(candidates, key=lambda x: x.sortkey, reverse=True)



import pkg_resources
from pkg_resources._vendor.packaging.version import Version


PART_MAX = '999999999'


def _offset_minor_version(version, offset, pos=2):
    parts = str(version).split('.')
    while len(parts) < 3:
        parts += ['0']

    try:
        cur_version = int(parts[pos])
        if cur_version == 0 and offset < 0:
            if pos == 0:
                raise ValueError('Cannot create a version less than 0')
            parts[pos] = PART_MAX
            return _offset_minor_version(pkg_resources.parse_version('.'.join(parts)), -1, pos=pos - 1)
        else:
            parts[pos] = str(int(parts[pos]) + offset)
        return pkg_resources.parse_version('.'.join(parts))
    except TypeError:
        return None


def is_possible(req):
    """
    Determine whether or not the requirement with its given specifiers is even possible.

    Args:
        req (Requirement):

    Returns:
        (bool) Whether or not the constraint can be satisfied
    """
    lower_bound = pkg_resources.parse_version('0.0.0')
    exact = None
    not_equal = []
    upper_bound = pkg_resources.parse_version('{max}.{max}.{max}'.format(max=PART_MAX))
    for spec in req.specifier:
        version = pkg_resources.parse_version(spec.version)
        if spec.operator == '==':
            if exact is None:
                exact = version
            if exact != version:
                return False
        if spec.operator == '!=':
            not_equal.append(version)
        elif spec.operator == '>':
            if version > lower_bound:
                lower_bound = _offset_minor_version(version, 1)
        elif spec.operator == '>=':
            if version >= lower_bound:
                lower_bound = version
        elif spec.operator == '<':
            if version < upper_bound:
                upper_bound = _offset_minor_version(version, -1)
        elif spec.operator == '<=':
            if version <= upper_bound:
                upper_bound = version
    if upper_bound < lower_bound:
        return False
    if exact is not None:
        if exact > upper_bound or exact < lower_bound:
            return False
    for check in not_equal:
        if check == exact:
            return False
    return True



import logging
import os

import pkg_resources

from qer import utils
import qer.repos.repository
from qer.repos.repository import Repository


class FindLinksRepository(Repository):
    def __init__(self, path, allow_prerelease=None):
        super(FindLinksRepository, self).__init__(allow_prerelease=allow_prerelease)
        self.path = path
        self._logger = logging.getLogger('qer.repository.findlinks')
        self.links = []
        self._find_all_links()

    def __repr__(self):
        return '--find-links {}'.format(self.path)

    def _find_all_links(self):
        for filename in os.listdir(self.path):
            candidate = qer.repos.repository.process_distribution(None, filename)
            if candidate is not None:
                self.links.append(candidate)

    @property
    def logger(self):
        return self._logger

    def get_candidates(self, req):
        project_name = utils.normalize_project_name(req.name)
        results = []
        for candidate in self.links:
            if utils.normalize_project_name(candidate.name) == project_name:
                results.append(candidate)

        return results

    def resolve_candidate(self, candidate):
        return os.path.join(self.path, candidate.filename), True

    def close(self):
        pass


if __name__ == '__main__':
    repo = FindLinksRepository('mywheels')
    print(repo.get_candidate(pkg_resources.Requirement.parse('astroid')))



from setuptools import setup, find_packages

setup(
    name='qer',
    version='0.6.1',
    author='Spencer Putt',
    author_email='sputt@alumni.iu.edu',
    description='Python requirements compiler',
    long_description=open('README.rst').read(),
    url='https://github.com/sputt/qer',
    install_requires=open('requirements.txt').readlines(),
    packages=find_packages(include=['qer*']),
    license='MIT License',
    entry_points={
        'console_scripts': [
            'req-compile = qer.cmdline:compile_main',
            'req-hash = qer.hash:hash_main',
            'req-candidates = qer.hash:candidates_main',
        ],
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Topic :: Software Development',
        'Intended Audience :: Developers',
    ],
)



import collections
import copy

import pkg_resources
import six

try:
    from functools32 import lru_cache
except ImportError:
    from functools import lru_cache

from qer import utils
from qer.utils import normalize_project_name, merge_requirements, filter_req, merge_extras


class ConstraintViolatedException(Exception):
    """Raised if a dist cannot be added"""
    def __init__(self, node):
        self.node = node


class DependencyNode(object):
    def __init__(self, key, req_name, metadata, extra=None):
        self.key = key
        self.metadata = metadata
        self.req_name = req_name
        self.extra = extra
        self.dependencies = {}  # Dict[DependencyNode, pkg_resources.Requirement]
        self.reverse_deps = set()  # Set[DependencyNode]

    def __repr__(self):
        return self.key

    def __str__(self):
        if self.metadata is None:
            return self.key + ' [UNSOLVED]'
        else:
            return '{}{}=={}'.format(
                self.metadata.name,
                ('[' + self.extra + ']') if self.extra else '',
                self.metadata.version)

    def build_constraints(self):
        req = None
        for node in self.reverse_deps:
            req = merge_requirements(req, node.dependencies[self])

        if req is None:
            if self.metadata is None:
                req = utils.parse_requirement(self.key)
            else:
                req = utils.parse_requirement(self.metadata.name)
            if self.extra:
                req.extras = (self.extra,)
                # Reparse to create a correct hash
                req = utils.parse_requirement(str(req))
        return req


class DistributionCollection(object):
    def __init__(self):
        self.nodes = {}  # Dict[str, DependencyNode]

    @staticmethod
    def _build_key(name, extra=None):
        return utils.normalize_project_name(name) + (('[' + extra + ']') if extra else '')

    def add_dist(self, metadata, source, reason):
        """
        Add a distribution

        Args:
            metadata (DistInfo, None): Distribution info to add
            source (DependencyNode, None): The source of the distribution, e.g. a filename
            reason (pkg_resources.Requirement, optional):
        """
        if reason is not None and len(reason.extras) > 1:
            for extra in reason.extras:
                new_req = copy.copy(reason)
                new_req.extras = (extra,)
                self.add_dist(metadata, source, new_req)
            return

        has_metadata = False
        if isinstance(metadata, six.string_types):
            req_name = metadata
        else:
            has_metadata = True
            req_name = metadata.name

        extra = reason.extras[0] if reason is not None and reason.extras else None
        key = DistributionCollection._build_key(req_name, extra)

        if key in self.nodes:
            node = self.nodes[key]
        else:
            node = DependencyNode(key, req_name, None, extra)
            self.nodes[key] = node

        if extra:
            # Add a reference back to the root req
            base_node = self.add_base(node, reason, req_name)
        else:
            base_node = node

        nodes = {base_node}
        if has_metadata:
            self.update_dists(base_node, metadata)

            # Apply the same metadata to all extras
            for reverse_node in base_node.reverse_deps:
                if reverse_node.req_name == req_name:
                    self.update_dists(reverse_node, metadata)
                    nodes.add(reverse_node)

        if base_node.metadata is not None and reason is not None:
            if not reason.specifier.contains(base_node.metadata.version):
                # Discard the metadata
                self.remove_dists(base_node, remove_upstream=False)

                for reverse_node in base_node.reverse_deps:
                    if reverse_node.req_name == req_name:
                        self.remove_dists(reverse_node, remove_upstream=False)

        if source is not None:
            node.reverse_deps.add(source)
            source.dependencies[node] = reason

        if not has_metadata:
            return set()
        return nodes

    def add_base(self, node, reason, req_name):
        if reason is not None:
            non_extra_req = copy.copy(reason)
            non_extra_req.extras = ()
            non_extra_req = utils.parse_requirement(str(non_extra_req))
        else:
            non_extra_req = utils.parse_requirement(req_name)

        self.add_dist(req_name, node, non_extra_req)
        return self.nodes[req_name]

    def update_dists(self, node, metadata):
        node.metadata = metadata
        for req in metadata.requires(node.extra):
            # This adds a placeholder entry
            self.add_dist(req.name, node, req)

    def remove_dists(self, node, remove_upstream=True):
        if isinstance(node, collections.Iterable):
            for single_node in node:
                self.remove_dists(single_node)
            return

        if node.key not in self.nodes:
            return

        if remove_upstream:
            del self.nodes[node.key]
            for reverse_dep in node.reverse_deps:
                del reverse_dep.dependencies[node]

        for dep in node.dependencies:
            if remove_upstream or dep.req_name != node.req_name:
                dep.reverse_deps.remove(node)
                if not dep.reverse_deps:
                    self.remove_dists(dep)

        if not remove_upstream:
            node.dependencies = {}
            node.metadata = None

    def build(self):
        results = []
        for node in self.nodes.values():
            if isinstance(node.metadata, DistInfo) and not node.extra:
                extras = []
                for reverse_dep in node.reverse_deps:
                    if reverse_dep.metadata.name == node.metadata.name:
                        extras.append(reverse_dep.extra)
                req_expr = '{}{}=={}'.format(
                    node.metadata.name,
                    ('[' + ','.join(sorted(extras)) + ']') if extras else '',
                    node.metadata.version)
                results.append(utils.parse_requirement(req_expr))
        return results

    def _build_constraints(self, root_node, exclude=None):
        constraints = []
        for node in root_node.reverse_deps:
            if node.req_name != exclude:
                req = node.dependencies[root_node]
                specifics = ' (' + str(req.specifier) + ')' if req.specifier else ''
                source = node.metadata.name + ('[' + node.extra + ']' if node.extra else '')
                constraints += [source + specifics]
        return constraints

    def generate_lines(self, filter=None):
        """
        Generate the lines of a results file from this collection
        Args:
            filter (Callable): Filter to apply to each element of the collection.
                Return True to keep a node, False to exclude it

        Returns:
            (list[str]) List of rendered node entries in the form of
                reqname==version   # reasons
        """
        filter = filter or (lambda _: True)
        for node in self.nodes.values():
            if isinstance(node.metadata, DistInfo) and not node.extra:
                extras = []
                constraints = self._build_constraints(node, exclude=node.metadata.name)
                for reverse_dep in node.reverse_deps:
                    if reverse_dep.metadata.name == node.metadata.name:
                        if reverse_dep.extra is None:
                            print('Reverse dep with none extra: {}'.format(reverse_dep))
                        extras.append(reverse_dep.extra)
                        constraints.extend(self._build_constraints(reverse_dep))
                try:
                    req_expr = '{}{}=={}'.format(
                        node.metadata.name,
                        ('[' + ','.join(sorted(extras)) + ']') if extras else '',
                        node.metadata.version)
                except TypeError:
                    print('Failed processing {}, extras={}'.format(node, extras))

                constraint_text = ', '.join(sorted(constraints))
                if filter(node):
                    if not node.metadata.meta:
                        yield '{}# {}'.format(req_expr.ljust(43), constraint_text)

    def __contains__(self, project_name):
        return normalize_project_name(project_name) in self.nodes

    def __iter__(self):
        return iter(self.nodes.values())

    def __getitem__(self, project_name):
        return self.nodes[normalize_project_name(project_name)]


class RequirementsFile(object):
    def __init__(self, filename, reqs):
        self.name = filename
        self.version = utils.parse_version('1')
        self.reqs = list(reqs)
        self.meta = True

    def requires(self, extra=None):
        return [req for req in self.reqs
                if filter_req(req, extra)]


class DistInfo(object):
    def __init__(self, name, version, reqs, meta=False):
        """
        Args:
            name:
            version:
            reqs:
            meta:
        """
        self.key = ''
        self.reqs = list(reqs)
        self.meta = meta
        self._version = version
        self._name = None
        self.name = name
        self._recalc_hash()
        self.source = None

    def __hash__(self):
        return self.hash

    def _recalc_hash(self):
        self.hash = hash(self.key + str(self.version))

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is not None:
            self._name = value
            self.key = utils.normalize_project_name(self._name)
            self._recalc_hash()

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = value
        self._recalc_hash()

    def requires(self, extra=None):
        return [req for req in self.reqs
                if filter_req(req, extra)]

    def __str__(self):
        return '{}=={}'.format(self.name, self.version)

    def __repr__(self):
        return self.name + ' ' + self.version + '\n' + '\n'.join([str(req) for req in self.reqs])



from __future__ import print_function

import functools
import io
import logging
import os
import sys
import tarfile
import tempfile
import types
import zipfile
import abc

import pkg_resources
import six
from six.moves import StringIO

from qer import utils
from qer.dists import DistInfo


class MetadataError(Exception):
    def __init__(self, name, version):
        self.name = name
        self.version = version


class Extractor(six.with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    def names(self):
        pass

    @abc.abstractmethod
    def open(self, filename, mode='r', encoding=None):
        pass

    @abc.abstractmethod
    def close(self):
        pass

    def relative_opener(self, fake_root, directory):
        def inner_opener(filename, *args, **kwargs):
            archive_path = filename
            if os.path.isabs(filename):
                if filename.startswith(fake_root):
                    archive_path = os.path.relpath(filename, fake_root)
                else:
                    return self.open(filename, *args, **kwargs)
            return self.open(directory + '/' + archive_path, *args, **kwargs)
        return inner_opener

    def contents(self, name):
        return self.open(name, encoding='utf-8').read()


def parse_source_filename(full_filename):
    filename = full_filename
    filename = filename.replace('.tar.gz', '')
    filename = filename.replace('.zip', '')
    filename = filename.replace('.tgz', '')

    dash_parts = filename.split('-')
    version_start = None
    for idx, part in enumerate(dash_parts):
        if part[0].isdigit():
            version_start = idx
            break

    if version_start is None:
        return os.path.basename(full_filename), None

    if version_start == 0:
        raise ValueError('Package name missing: {}'.format(full_filename))

    pkg_name = '-'.join(dash_parts[:version_start])
    version = utils.parse_version('-'.join(dash_parts[version_start:]))

    return pkg_name, version


class NonExtractor(Extractor):
    def __init__(self, path):
        self.path = path
        self.io_open = io.open

    def names(self):
        parent_dir = os.path.abspath(os.path.join(self.path, '..'))
        for root, dirs, files in os.walk(self.path):
            rel_root = os.path.relpath(root, parent_dir).replace('\\', '/')
            for file in files:
                yield rel_root + '/' + file

    def open(self, filename, mode='r', encoding='utf-8', errors=None):
        if not os.path.isabs(filename):
            parent_dir = os.path.abspath(os.path.join(self.path, '..'))
            return self.io_open(os.path.join(parent_dir, filename), mode=mode, encoding=encoding)
        else:
            return self.io_open(filename, mode=mode, encoding=encoding)

    def close(self):
        pass


class TarExtractor(Extractor):
    def __init__(self, filename):
        self.tar = tarfile.open(filename, 'r:gz')
        self.io_open = io.open

    def names(self):
        return (info.name for info in self.tar.getmembers())

    def open(self, filename, mode='r', encoding='utf-8', errors=None):
        filename = filename.replace('\\', '/').replace('./', '')
        if not os.path.isabs(filename):
            try:
                handle = self.tar.extractfile(filename)
                return with_decoding(handle, encoding=encoding if mode != 'rb' else None)
            except KeyError:
                raise IOError('Not found in archive: {}'.format(filename))
        else:
            return self.io_open(filename, mode=mode, encoding=encoding)

    def close(self):
        self.tar.close()


class ZipExtractor(Extractor):
    def __init__(self, filename):
        self.zfile = zipfile.ZipFile(filename, 'r')
        self.io_open = io.open

    def names(self):
        return self.zfile.namelist()

    def open(self, filename, mode='r', encoding='utf-8', errors=None):
        filename = filename.replace('\\', '/').replace('./', '')
        if not os.path.isabs(filename):
            try:
                output = with_decoding(StringIO(self.zfile.read(filename).decode(encoding)), None)
                return output
            except KeyError:
                raise IOError('Not found in archive: {}'.format(filename))
        else:
            return self.io_open(filename, mode=mode, encoding=encoding)

    def close(self):
        self.zfile.close()


def extract_metadata(dist, origin=None):
    """"""
    if dist.lower().endswith('.whl'):
        result = _fetch_from_wheel(dist)
    elif dist.lower().endswith('.zip'):
        result = _fetch_from_source(dist, ZipExtractor)
    elif dist.lower().endswith('.tar.gz'):
        result = _fetch_from_source(dist, TarExtractor)
    else:
        result = _fetch_from_source(os.path.abspath(dist), NonExtractor)
    if result is not None:
        result.origin = origin
    return result


def _fetch_from_source(source_file, extractor_type):
    """

    Args:
        source_file (str): Source file
        extractor_type (type[Extractor]): Type of extractor to use

    Returns:

    """
    extractor = extractor_type(source_file)  # type: Extractor
    filename = os.path.basename(source_file)
    name, version = parse_source_filename(filename)
    try:
        metadata_file = None
        pkg_info_file = None
        egg_info = None
        setup_file = None

        for info_name in extractor.names():
            if info_name.lower().endswith('pkg-info') and info_name.count('/') <= 1:
                pkg_info_file = info_name
            elif info_name.endswith('.egg-info/requires.txt'):
                egg_info = info_name
            elif info_name.endswith('metadata') and info_name.count('/') <= 1:
                metadata_file = info_name

            if info_name.endswith('setup.py') and info_name.count('/') <= 1:
                setup_file = info_name
                break

        results = None
        if egg_info:
            requires_contents = ''
            try:
                requires_contents = extractor.open(egg_info, encoding='utf-8').read()
            except KeyError:
                pass
            return _parse_requires_file(requires_contents,
                                        name,
                                        version)

        if pkg_info_file:
            results = _parse_flat_metadata(extractor.open(pkg_info_file, encoding='utf-8').read())

        if (results is None or not results.reqs) and setup_file:
            if isinstance(extractor, NonExtractor):
                fake_setupdir = source_file
            else:
                fake_setupdir = tempfile.mkdtemp()
            setup_results = _parse_setup_py(name, version, fake_setupdir,
                                            extractor.relative_opener(fake_setupdir,
                                                                      os.path.dirname(setup_file)))
            if setup_results is not None:
                if version is not None:
                    setup_results.version = version
                if setup_results.name is None:
                    setup_results.name = name
                if results:
                    setup_results.version = results.version
                return setup_results
            elif results is not None:
                return results
            else:
                return None

        if metadata_file:
            return _parse_flat_metadata(extractor.open(metadata_file, encoding='utf-8').read())
    finally:
        extractor.close()


def _fetch_from_wheel(wheel):
    zfile = zipfile.ZipFile(wheel, 'r')
    try:
        metadata_file = None
        infos = zfile.namelist()
        for info in infos:
            if info.lower().endswith('metadata'):
                metadata_file = info

        if metadata_file:
            return _parse_flat_metadata(zfile.read(metadata_file).decode('utf-8'))
    finally:
        zfile.close()


def _parse_flat_metadata(contents):
    name = None
    version = None
    raw_reqs = []

    for line in contents.split('\n'):
        if line.lower().startswith('name:'):
            name = line.split(':')[1].strip()
        if line.lower().startswith('version:'):
            version = utils.parse_version(line.split(':')[1].strip())
        if line.lower().startswith('requires-dist:'):
            raw_reqs.append(line.split(':')[1].strip())

    return DistInfo(name, version, list(utils.parse_requirements(raw_reqs)))


class with_decoding(object):
    def __init__(self, wrap, encoding):
        self.file = wrap
        self.encoding = encoding

    def read(self):
        results = self.file.read()
        if self.encoding:
            results = results.decode(self.encoding)
        if six.PY2:
            results = str(''.join([i if ord(i) < 128 else ' ' for i in results]))
        return results

    def readlines(self):
        results = self.file.readlines()
        if self.encoding:
            results = [result.decode(self.encoding) for result in results]
        return results

    def write(self, *args, **kwargs):
        pass

    def __iter__(self):
        if self.encoding:
            return (line.decode(self.encoding) for line in self.file)
        return iter(self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def close(self):
        pass


def setup(results, *args, **kwargs):
    name = kwargs.get('name', None)
    version = kwargs.get('version', '0.0.0')
    reqs = kwargs.get('install_requires', [])
    extra_reqs = kwargs.get('extra_requires', {})

    if version is None or isinstance(version, FakeModule):
        version = '0.0.0'

    if version is not None and version != '':
        version = pkg_resources.parse_version(str(version))

    all_reqs = list(utils.parse_requirements(reqs))
    for extra, extra_req_strs in extra_reqs.items():
        cur_reqs = [utils.parse_requirement('{} ; extra=="{}"'.format(reqstr, extra))
                    for reqstr in extra_req_strs]
        all_reqs.extend(cur_reqs)

    results.append(DistInfo(name, version, all_reqs))
    return FakeModule('dist')


class FakeModule(types.ModuleType):
    call_count = 0

    def __getitem__(self, item):
        self.call_count += 1
        if self.call_count > 30:
            raise ValueError('Too many')
        return None

    def __contains__(self, item):
        return True

    def __call__(self, *args, **kwargs):
        return FakeModule(self.__name__)

    def __setitem__(self, key, value):
        pass

    def __getattr__(self, item):
        if isinstance(item, str):
            if item == '__path__':
                return ''
            elif item == '__file__':
                return os.path.join(self.__getattribute__('__name__'), '__init__.py')
            return FakeModule(item)
        else:
            return None


def fake_import(name, orig_import, modname, *args, **kwargs):
    try:
        lower_modname = modname.lower()
        if 'build_ext' in modname:
            raise ImportError()

        if 'Cython' in modname or (
                name == lower_modname or (name + '_') in lower_modname or
                ('_' + name) in lower_modname):
            # print('Module is fake: {}'.format(modname))
            sys.modules[modname] = FakeModule(modname)
        return orig_import(modname, *args, **kwargs)
    except ImportError:
        # print('Failed on {}'.format(modname))
        # Skip any cython importing to improve setup.py compatibility (e.g. subprocess32)
        if 'Cython' in modname:
            raise
        if six.PY2:
            if 'asyncio' in modname:
                raise

        #     if 'typing' in modname:
        #         raise
        #
        if name in modname.lower() or '_version' in modname or 'version' in modname:
            modparts = modname.split('.')
            for idx, mod in enumerate(modparts):
                sys.modules['.'.join(modparts[:idx + 1])] = FakeModule(mod)
        # print('Failed, original import')
        # try:
        #     return orig_import(modname, *args, **kwargs)
        # except TypeError:
        #     pass
        return orig_import(modname, *args, **kwargs)
    finally:
        sys.stdout.flush()


def _parse_setup_py(name, version, fake_setupdir, opener):
    import setuptools
    import distutils.core
    import sys

    # Capture warnings.warn, which is sometimes used in setup.py files
    logging.captureWarnings(True)

    sys.exit = lambda code: None

    results = []
    setup_with_results = functools.partial(setup, results)
    if six.PY2:
        import __builtin__
        old_import = __builtin__.__import__
        __builtin__.__import__ = functools.partial(fake_import, name.lower(), __import__)
    else:
        import builtins
        old_import = builtins.__import__
        builtins.__import__ = functools.partial(fake_import, name.lower(), __import__)

    import imp
    old_load_source = imp.load_source
    imp.load_source = lambda *args, **kwargs: FakeModule('load_source')

    import codecs
    old_codecs_open = codecs.open
    codecs.open = opener

    old_stdout = sys.stdout
    old_stderr = sys.stderr

    sys.stderr = StringIO()
    sys.stdout = StringIO()

    old_listdir = os.listdir
    os.listdir = lambda path: []

    curr_dir = os.getcwd()

    old_open = io.open
    io.open = opener

    spy_globals = {'__file__': os.path.join(fake_setupdir, 'setup.py'),
                   '__name__': '__main__',
                   'open': opener,
                   'setup': setup_with_results}

    setuptools.setup = setup_with_results
    distutils.core.setup = setup_with_results
    contents = opener('setup.py', encoding='utf-8').read()
    try:
        if six.PY2:
            lines = contents.split('\n')
            lines = [line for line in lines if not (line.startswith('#') and
                                                    ('-*- coding' in line or '-*- encoding' in line or 'encoding:' in line))]
            contents = '\n'.join(lines)
        contents = contents.replace('print ', '')
        exec(contents, spy_globals, spy_globals)
    except:
        raise MetadataError(name, version)
    finally:
        os.chdir(curr_dir)
        io.open = old_open
        if six.PY2:
            __builtin__.__import__ = old_import
        else:
            builtins.__import__ = old_import
        imp.load_source = old_load_source
        codecs.open = old_codecs_open

        sys.stderr = old_stderr
        sys.stdout = old_stdout

        os.listdir = old_listdir
    if not results:
        pass
    return results[0]


def _parse_requires_file(contents, name, version):
    reqs = []
    sections = list(pkg_resources.split_sections(contents))
    for section in sections:
        if section[0] is None:
            reqs.extend(utils.parse_requirements(section[1]))
        elif section[0].startswith(':python_version'):
            for req in section[1]:
                reqs.append(utils.parse_requirement(req + ' ' + section[0].replace(':', ';')))

    return DistInfo(name, version, reqs)



from __future__ import print_function

import argparse
import logging
import os
import shutil
import sys
import tempfile

import qer.compile
import qer.dists
import qer.metadata
import qer.repos.pypi

from qer import utils
from qer.compile import perform_compile
from qer.repos.findlinks import FindLinksRepository
from qer.repos.pypi import PyPIRepository
from qer.repos.repository import CantUseReason
from qer.repos.multi import MultiRepository
from qer.repos.solution import SolutionRepository
from qer.repos.source import SourceRepository
from qer.versions import is_possible


def _cantusereason_to_text(reason):
    if reason == CantUseReason.VERSION_NO_SATISFY:
        return 'version mismatch'
    elif reason == CantUseReason.WRONG_PLATFORM:
        return 'platform mismatch {}'.format(qer.repos.repository.PLATFORM_TAG)
    elif reason == CantUseReason.WRONG_PYTHON_VERSION:
        return 'python version/interpreter mismatch ({})'.format(', '.join(
            qer.repos.repository.RequiresPython.WHEEL_VERSION_TAGS))
    elif reason == CantUseReason.IS_PRERELEASE:
        return 'prereleases not used'

    return 'unknown'


def _generate_no_candidate_display(ex, repo, dists):
    if isinstance(repo, MultiRepository):
        repos = repo.repositories
    else:
        repos = [repo]

    failing_node = dists[ex.req.name]
    nodes = failing_node.reverse_deps
    constraints = failing_node.build_constraints()
    if not nodes:
        print('No package available for {}, latest is {}'.format(ex.req.name, 'unknown'), file=sys.stderr)
    else:
        can_satisfy = is_possible(constraints)
        if not can_satisfy:
            print('No version could possibly satisfy the following requirements:', file=sys.stderr)
        else:
            print('No version of {} could satisfy the following requirements:'.format(ex.req.name), file=sys.stderr)

        for node in nodes:
            print('   {} requires {}'.format(node, node.dependencies[failing_node]), file=sys.stderr)

        if can_satisfy:
            all_candidates = {repo: repo.get_candidates(ex.req) for repo in repos}
            if sum(len(candidates) for candidates in all_candidates.values()) == 0:
                print('No candidates found in any of the input sources', file=sys.stderr)
            else:
                print('Found the following candidates, none of which will work:', file=sys.stderr)
                for repo in repos:
                    candidates = repo.get_candidates(ex.req)
                    print('  {}:'.format(repo), file=sys.stderr)
                    if candidates:
                        for candidate in repo._sort_candidates(candidates):
                            print('  {}: {}'.format(candidate,
                                                    _cantusereason_to_text(repo.why_cant_I_use(ex.req, candidate))),
                                  file=sys.stderr)
                    else:
                        print('  No candidates found', file=sys.stderr)


def run_compile(input_reqfiles, constraint_files, source, force_extras, find_links,
                index_url, wheeldir, no_index, remove_source, solution):

    if wheeldir:
        if not os.path.exists(wheeldir):
            os.mkdir(wheeldir)
        delete_wheeldir = False
    else:
        wheeldir = tempfile.mkdtemp()
        delete_wheeldir = True

    if len(input_reqfiles) == 1 and os.path.isdir(input_reqfiles[0]):
        dist = qer.metadata.extract_metadata(input_reqfiles[0])
        input_reqs = [utils.parse_requirement(dist.name)]
    else:
        input_reqs = {
            req_file: utils.reqs_from_files([req_file])
            for req_file in input_reqfiles
        }

    if constraint_files:
        constraint_reqs = utils.reqs_from_files(constraint_files)
    else:
        constraint_reqs = None

    repo = build_repo(solution, source, force_extras, find_links, index_url, no_index, wheeldir)

    try:
        results = perform_compile(input_reqs, repo, constraint_reqs=constraint_reqs)

        filter = None

        if remove_source:
            source_repo = None
            for repo in repo.repositories:
                if isinstance(repo, SourceRepository):
                    source_repo = repo

            if source_repo is None:
                raise ValueError('Cannot remove results from source, no source provided')

            filter = lambda dist: not hasattr(dist.metadata, 'origin') or dist.metadata.origin is not source_repo

        lines = sorted(results.generate_lines(filter=filter), key=str.lower)
        print('\n'.join(lines))
    except qer.repos.repository.NoCandidateException as ex:
        _generate_no_candidate_display(ex, repo, ex.results)
        sys.exit(1)

    if delete_wheeldir:
        shutil.rmtree(wheeldir)


def build_repo(solution, source, force_extras, find_links, index_url, no_index, wheeldir):
    repos = []
    if solution:
        repos.append(SolutionRepository(solution))
    if source:
        repos.append(SourceRepository(source, force_extras=force_extras))
    if find_links:
        repos.append(FindLinksRepository(find_links))
    if not no_index:
        index_url = index_url or 'https://pypi.org/simple'
        repos.append(PyPIRepository(index_url, wheeldir))
    if not repos:
        raise ValueError('At least one Python distributions source must be provided.')
    if len(repos) > 1:
        repo = MultiRepository(*repos)
    else:
        repo = repos[0]
    return repo


def compile_main():
    logging.basicConfig(level=logging.ERROR)

    parser = argparse.ArgumentParser()
    parser.add_argument('requirement_files', nargs='+', help='Input requirements files')
    parser.add_argument('-c', '--constraints', action='append',
                        help='Contraints files. Not included in final compilation')
    parser.add_argument('--remove-source', default=False, action='store_true',
                        help='Remove distributions satisfied via --source from the output')
    parser.add_argument('-e', '--extra', nargs='+', default=[],
                        help='Extras to include for every discovered distribution '
                             'provided via --source')
    parser.add_argument('-u', '--solution', type=str, default=None,
                        help='Existing fully-pinned constraints file to use as a baseline when compiling')

    add_repo_args(parser)

    args = parser.parse_args()
    run_compile(args.requirement_files, args.constraints if args.constraints else None, args.source, tuple(args.extra), args.find_links, args.index_url,
                args.wheel_dir, args.no_index, args.remove_source, args.solution)


def add_repo_args(parser):
    parser.add_argument('-i', '--index-url', type=str, default=None)
    parser.add_argument('-f', '--find-links', type=str, default=None)
    parser.add_argument('-s', '--source', type=str, default=None)
    parser.add_argument('-w', '--wheel-dir', type=str, default=None)
    parser.add_argument('--no-index',
                        action='store_true', default=False,
                        help='Do not connect to the internet to compile. All wheels must be '
                             'available in --find-links paths.')


if __name__ == '__main__':
    compile_main()



from __future__ import print_function
import collections
import logging
import os

from qer import utils
import qer.metadata
import qer.repos.repository

from qer.repos.repository import Repository


SPECIAL_DIRS = ('site-packages', 'dist-packages', '.git', '.svn', '.idea')
SPECIAL_FILES  = ('__init__.py',)


class SourceRepository(Repository):
    def __init__(self, path, allow_prerelease=None, force_extras=()):
        super(SourceRepository, self).__init__(allow_prerelease=allow_prerelease)
        self.path = path
        self._logger = logging.getLogger('qer.repository.source')
        self.distributions = collections.defaultdict(list)
        self._find_all_distributions()
        self._force_extras = force_extras

    def _find_all_distributions(self):
        for root, dirs, files in os.walk(self.path):
            for dir_ in dirs:
                if dir_ in SPECIAL_DIRS:
                    dirs.remove(dir_)

            for file in files:
                if file in SPECIAL_FILES:
                    for dir_ in dirs:
                        dirs.remove(dir_)
                elif file == 'setup.py':
                    for dir_ in list(dirs):
                        dirs.remove(dir_)

                    result = qer.metadata.extract_metadata(root)
                    candidate = qer.repos.repository.Candidate(
                        result.name,
                        root,
                        result.version,
                        qer.repos.repository.RequiresPython(None),
                        'any',
                        None,
                        qer.repos.repository.DistributionType.SOURCE)
                    candidate.preparsed = result
                    self.distributions[utils.normalize_project_name(result.name)].append(candidate)
                    break

    def __repr__(self):
        return '--source {}'.format(self.path)

    @property
    def logger(self):
        return self._logger

    def get_candidates(self, req):
        project_name = utils.normalize_project_name(req.name)
        return self.distributions.get(project_name, [])

    def resolve_candidate(self, candidate):
        return candidate.filename, True

    def force_extras(self):
        return self._force_extras

    def close(self):
        pass
